<?php

namespace Drupal\cfrapi\Configurator\Optional;

use Drupal\cfrapi\CfrCodegenHelper\CfrCodegenHelperInterface;
use Drupal\cfrapi\ConfEmptyness\ConfEmptynessInterface;
use Drupal\cfrapi\Exception\ConfToValueException;
use Drupal\cfrapi\SummaryBuilder\SummaryBuilderInterface;

abstract class OptionalConfiguratorBase implements OptionalConfiguratorInterface, ConfEmptynessInterface {

  /**
   * @var bool
   */
  private $required;

  /**
   * @param $required
   */
  public function __construct($required = TRUE) {
    $this->required = $required;
  }

  /**
   * @return bool
   */
  protected function isRequired() {
    return $this->required;
  }

  /**
   * {@inheritdoc}
   */
  public function confGetSummary($conf, SummaryBuilderInterface $summaryBuilder) {

    if ($this->confIsEmpty($conf)) {
      if ($this->required) {
        return '- ' . t('Missing') . ' -';
      }

      return '- ' . t('None') . ' -';
    }

    return $this->nonEmptyConfGetSummary($conf, $summaryBuilder);
  }

  /**
   * Gets a summary for a non-empty configuration.
   *
   * (This method will not be called with empty configuration)
   *
   * @param mixed $conf
   *   Non-empty configuration.
   * @param \Drupal\cfrapi\SummaryBuilder\SummaryBuilderInterface $summaryBuilder
   *   Helper object for summary-building.
   *
   * @return mixed
   *   The summary, usually a string of html.
   */
  abstract protected function nonEmptyConfGetSummary($conf, SummaryBuilderInterface $summaryBuilder);

  /**
   * {@inheritdoc}
   */
  public function confGetValue($conf) {

    if ($this->confIsEmpty($conf)) {
      if ($this->required) {
        throw new ConfToValueException("Required, but empty.");
      }

      return NULL;
    }

    return $this->nonEmptyConfGetValue($conf);
  }

  /**
   * {@inheritdoc}
   */
  public function confGetPhp($conf, CfrCodegenHelperInterface $helper) {

    if ($this->confIsEmpty($conf)) {
      if ($this->required) {
        return $helper->incompatibleConfiguration($conf, "Required, but empty.");
      }

      return 'NULL';
    }

    return $this->nonEmptyConfGetPhp($conf, $helper);
  }

  /**
   * Gets a value for non-empty configuration.
   *
   * (This method will not be called with empty configuration)
   *
   * @param mixed $conf
   *   Non-empty configuration.
   *
   * @return mixed
   *   Value for the given configuration.
   *
   * @throws \Drupal\cfrapi\Exception\ConfToValueException
   *   If no value can be produced for the given configuration.
   */
  abstract protected function nonEmptyConfGetValue($conf);

  /**
   * @param mixed $conf
   * @param \Drupal\cfrapi\CfrCodegenHelper\CfrCodegenHelperInterface $helper
   *
   * @return mixed
   */
  protected function nonEmptyConfGetPhp($conf, CfrCodegenHelperInterface $helper) {
    return $helper->notSupported($this, $conf, "nonEmptyConfGetPhp() not supported.");
  }

  /**
   * {@inheritdoc}
   */
  public function getEmptyness() {
    return $this->required ? NULL : $this;
  }

  /**
   * {@inheritdoc}
   */
  public function confIsEmpty($conf) {
    return NULL === $conf || '' === $conf || [] === $conf;
  }

  /**
   * {@inheritdoc}
   */
  public function getEmptyConf() {
    return NULL;
  }

}
